#! /usr/bin/perl
#**********************************************************************
#* (C) Copyright IBM Corporation 2004, 2004.  All rights reserved.    *
#>                                                                    *
#* Tivoli Storage Manager                                             *
#*                                                                    *
#* This script defines all registered nodes in domain_name that are   *
#* not currently defined as members of a collocation group to new     *
#* groups based on the amount of data currently stored under those    *
#* nodes and the specified media capacity.  The new collocation       *
#* groups are created with names using the group_prefix prefix and a  *
#* sequentially increasing numeric suffix.  The capacity is to be     *
#* specified in MB and is used only by this script, which does not    *
#* limit the storage capacity on the TSM server for these nodes in    *
#* any manner.  A recommended value to be used for capacity is 4      *
#* times the expected or realized sequential media volume capacity    *
#* (which might be determined using the fullvolcapacity script).      *
#* This value will likely provide a favorable balance so that the     *
#* storage volumes are filled and data for any given node is spread   *
#* across an optimal number of storage volumes for multi-session      *
#* restore.  Collocation by group will only be used for data that is  *
#* stored in sequential storage pools that are configured with        *
#* collocate=group.                                                   *
#*                                                                    *
#* Collocation groups are only supported by TSM server version 5.3    *
#* or higher, however, this script may be run on TSM servers prior    *
#* to 5.3.  In that case, all nodes in the domain are included, and   *
#* the collocation groups are not defined (regardless of -EXEC).      *
#*                                                                    *
#* Usage: defgroups -ID=userid -PAssword=password -EXEC               *
#*                  domain_name group_prefix capacity                 *
#*                                                                    *
#*        where -ID=userid specifies the administrator's user ID      *
#*              -PAssword=password specifies the administrator's      *
#*                 password                                           *
#*              -EXEC specifies that the new collocation groups will  *
#*                 be defined with their members as determined by     *
#*                 this script.  If this argument is not specified,   *
#*                 information about the groups that would be defined *
#*                 is displayed, and a macro called _defgrps.mac is   *
#*                 created in the current directory, but the macro is *
#*                 not executed and the groups are not defined.       *
#*              domain_name specifies the TSM policy domain name      *
#*              group_prefix specifies the prefix to be used for      *
#*                 creating the collocation group names               *
#*              capacity specifies the amount of data in MB to be     *
#*                 used for creating the collocation groups and       *
#*                 defining nodes to them                             *
#*                                                                    *
#* Dependencies: UNIX or Windows                                      *
#*               Perl interpreter installed                           *
#*               TSM administrative client installed and configured   *
#>                                                                    *
#* Note: This sample script is shipped "AS-IS" with all TSM server    *
#* platforms and must remain formatted with lines less than 80        *
#* characters each.                                                   *
#*                                                                    *
#**********************************************************************
use strict;
if ( @ARGV == 0 ) { usage() }

my $Id = "";
my $Pa = "";
my $Ex = 0;
while ( substr( $ARGV[0], 0, 1 ) eq "-" ) {
   my $A = uc( $ARGV[0] );
   if ( $A eq "-EXEC" )                { $Ex = 1; shift; next; }
   if ( chkarg( $A, "-ID", 3 ) )       { $Id = $ARGV[0]; shift; next; }
   if ( chkarg( $A, "-PASSWORD", 3 ) ) { $Pa = $ARGV[0]; shift; next; }
   if ( $ARGV[0] eq "-h" ) { usage() }
   print "Syntax error: Argument $ARGV[0] not understood.\n"; shift;
}

if ( $Id eq '' ) {
   print "Syntax error: TSM administrator's user ID must be supplied.\n";
   usage();
   exit;
}
if ( $Pa eq '' ) {
   print "Syntax error: TSM administrator's password must be supplied.\n";
   usage();
   exit;
}
if ( @ARGV < 3 ) {
   print "Syntax error: Missing required arguments.\n";
   usage();
   exit;
}
if ( @ARGV > 3 ) {
   print "Syntax error: Extraneous arguments supplied.\n";
   usage();
   exit;
}

my $Domain = uc( $ARGV[0] );
my $Prefix = uc( $ARGV[1] );
my $Capacity = $ARGV[2];

if ( $Capacity < 500 ) {
   print "Argument error: Capacity must be at least 500 MB.\n";
   exit;
}

my $Macro  = "_defgrps.mac";
my $CurrentNode = 0;
my $CurrentGroup = 1;
my @Nodes  = ();
my @Groups = ();
my $Ver = 0;

getver();
getnodes();
if ( @Nodes == 0 ) {
   print "\nThere are no registered nodes in policy domain '$Domain' ";
   print "that need to be added to a collocation group.\n";
   exit;
}

getgrps();
defbgrps();
preagrps();
defnodes();

if ( $Ex == 1 ) { print "\nCollocation groups defined:\n\n"; }
else            {
   print "\nCollocation groups that would be defined:\n\n";
}
print "                                 Number of nodes Number of MBs of\n";
print "                                 defined to      physical space used\n";
print "Group name                       this group      by these nodes\n";
print "-------------------------------- --------------- -------------------\n";

my $I;
for ($I = 0; $I < @Groups; $I++) {
   if ( $Groups[$I][1] == 1 ) {
      printf "%-32s %15s %19.0d\n", $Groups[$I][0],
         $Groups[$I][2], $Groups[$I][3];
   }
}

exit;

sub usage
#**********************************************************************
#* Display program usage information by reading program code header   *
#*                                                                    *
#**********************************************************************
{
   my $Skip = 1;
   open USAGE, "<$0";
   while ( <USAGE> ) {
      if ( substr( $_, 0, 2 ) eq "#>" ) {
         if ( $Skip == 1 ) { $Skip = 0 }
         else              { last }
      }
      if ( $Skip == 0 ) { chop; chop; print substr( $_, 3 )."\n" }
   }
   close USAGE;
   exit 4;
}

sub chkarg
#**********************************************************************
#* Check supplied userid and password arguments                       *
#*                                                                    *
#**********************************************************************
{
   my ( $Arg, $Flag, $Length ) = @_;
   my @Arg = split "=", $Arg;
   if ( length( $Arg[0] ) > $Length ) { $Length = length( $Arg[0] ) }
   if ( $Arg[0] eq substr( $Flag, 0, $Length ) and $Arg[1] ne '' ) {
      return 1;
   }
   else {
      return 0;
   }
}

sub getver
#**********************************************************************
#* Get the current server version.                                    *
#*                                                                    *
#**********************************************************************
{
   my ( $I, $J, $L, @L, @Lines );

   @Lines = split "\n", `dsmadmc $Id $Pa -noconfirm query status`;
   for ( $I = 0; $I < @Lines; $I++ ) {
      $J = index( uc( $Lines[$I] ), " VERSION " );
      if ( $J >= 0 ) {
         $L = substr( $Lines[$I], $J + 2 );
         $L =~ s/,//g;
         @L = split " ", $L;
         $Ver = $L[1].".".$L[3];
      }
   }
   if ( $Ver < 5.3 ) { $Ex = 0 }          # Cannot define colloc groups
}

sub getnodes
#**********************************************************************
#* Get the list of nodes not currently defined as members of a        *
#* collocation group and the amount of primary storage pool space     *
#* they are using.                                                    *
#*                                                                    *
#**********************************************************************
{
   my ( $I, @L, @Lines, $Opts );

   open  MACRO, ">$Macro" or die;
   print MACRO  "select '-', nodes.node_name,";
   print MACRO  " sum(physical_mb) from nodes,occupancy";
   print MACRO  " where nodes.domain_name='$Domain' and";

   if ( $Ver >= 5.3 ) {
      print MACRO " nodes.collocgroup_name is null and";
   }

   print MACRO  " nodes.node_name=occupancy.node_name and";
   print MACRO  " stgpool_name in ( select stgpool_name";
   print MACRO  " from stgpools where pooltype='PRIMARY' )";
   print MACRO  " group by nodes.node_name order by 3 desc\n";
   close MACRO;

   print "\nExecuting an SQL query which may ";
   print "require a significant amount of time...\n";
   $Opts = "$Id $Pa -noconfirm -commadelimited";
   @Lines = split "\n", `dsmadmc $Opts macro $Macro`;

   for ( $I = 0; $I < @Lines; $I++ ) {
      if ( substr( $Lines[$I], 0, 2 ) eq "-," ) {
         @L = split ",", $Lines[$I];
         push @Nodes, [];
         $Nodes[-1][0] = $L[1];                             # Node name
         $Nodes[-1][1] = $L[2];              # Node used physical space
      }
   }

   print "Finished SQL query.\n";
   unlink "$Macro";
}

sub getgrps
#**********************************************************************
#* Get a list of currently defined collocation groups with this       *
#* prefix.                                                            *
#*                                                                    *
#**********************************************************************
{
   my ( $I, @Lines, $Opts );

   open  MACRO, ">$Macro" or die;
   print MACRO  "select '-', collocgroup_name from collocgroup";
   print MACRO  " where collocgroup_name like '$Prefix"."%'\n";
   close MACRO;

   $Opts = "$Id $Pa -noconfirm -commadelimited";
   @Lines = split "\n", `dsmadmc $Opts macro $Macro`;

   for ( $I = 0; $I < @Lines; $I++ ) {
      if ( substr( $Lines[$I], 0, 2 ) eq "-," ) {
         push @Groups, [];
         $Groups[-1][0] = substr( $Lines[$I], 2 );         # Group name
         $Groups[-1][1] = 0;                               # New group?
         $Groups[-1][2] = 0;                       # No. of nodes added
         $Groups[-1][3] = 0;                          # No. MB of space
      }
   }

   unlink "$Macro";
}

sub getagrp
#**********************************************************************
#* Get a new unused collocation group name with the given prefix.     *
#*                                                                    *
#**********************************************************************
{
   my ( $G, $I );

   while( 1 ) {
      $G = $Prefix.$CurrentGroup;
      for ( $I = 0; $I < @Groups; $I++ ) {
         if ( $G eq $Groups[$I][0] ) { last }
      }
      if ( $I < @Groups ) {              # Group name exists, try again
         $CurrentGroup++;
      }
      else {
         push @Groups, [];                       # New group name found
         $Groups[-1][0] = $G;
         $Groups[-1][1] = 1;
         $Groups[-1][2] = 0;
         $Groups[-1][3] = 0;
         $CurrentGroup++;
         last;
      }
   }
   return $G;
}

sub updgrp
#**********************************************************************
#* Update the statistics for a new collocation group.                 *
#* Call with the group name, number of nodes defined to it, and the   *
#* number of MB of physical space occupied by those nodes.            *
#*                                                                    *
#**********************************************************************
{
   my ( $I );
   my ( $G, $N, $M ) = @_;

   for ( $I = 0; $I < @Groups; $I++ ) {
      if ( $G eq $Groups[$I][0] ) { last }
   }
   if ( $I < @Groups ) {
      $Groups[$I][2] += $N;
      $Groups[$I][3] += $M;
   }
}

sub defbgrps
#**********************************************************************
#* Define a single collocation group for each node with used space    *
#* that exceeds the specified capacity.                               *
#*                                                                    *
#**********************************************************************
{
   my ( $G );

   open MACRO, ">$Macro" or die;                          # New file!!!
   for ( $CurrentNode = 0; $CurrentNode < @Nodes; $CurrentNode++ ) {
      if ( $Nodes[$CurrentNode][1] > $Capacity ) {
         $G = getagrp();
         print MACRO "define collocgroup $G\n";
         print MACRO "define collocmember $G $Nodes[$CurrentNode][0]\n";
         updgrp( $G, 1, $Nodes[$CurrentNode][1] );
      }
      else { last }
   }
   close MACRO;
}

sub preagrps
#**********************************************************************
#* Pre-allocate a set of collocation groups for the nodes with used   *
#* space less than the specified capacity.                            *
#*                                                                    *
#**********************************************************************
{
   my ( $G, $I, $Ngroups, $Total );

   $Total = 0;
   for( $I = $CurrentNode; $I < @Nodes; $I++ ) {
      $Total += $Nodes[$I][1];
   }

   $Ngroups = int( 1 + ( $Total / $Capacity ) );

   open MACRO, ">>$Macro" or die;                           # Append!!!
   for ( $I = 0; $I < $Ngroups; $I++ ) {
      $G = getagrp();
      print MACRO "define collocgroup $G\n";
   }
   close MACRO;
}

sub defnodes
#**********************************************************************
#* Define the nodes with used space less than the specified capacity  *
#* to the pre-allocated collocation groups.                           *
#*                                                                    *
#**********************************************************************
{
   my ( $G, $I, $J, $K );

   my $NextGroup = 0;
   open MACRO, ">>$Macro" or die;                           # Append!!!
   for( $I = $CurrentNode; $I < @Nodes; $I++ ) {
      for( $J = 0; $J < @Groups; $J++ ) {
         $K = ( $NextGroup + $J ) % @Groups;
         if ( $Groups[$K][1] == 1 and
              ( $Groups[$K][3] + $Nodes[$I][1] < $Capacity ) ) {
            print MACRO "define collocmember $Groups[$K][0] $Nodes[$I][0]\n";
            updgrp( $Groups[$K][0], 1, $Nodes[$I][1] );
            $NextGroup = ( $K + 1 ) % @Groups;
            last;
         }
      }
      if ( $J == @Groups ) {
         $G = getagrp();
         print MACRO "define collocgroup $G\n";
         print MACRO "define collocmember $G $Nodes[$I][0]\n";
         updgrp( $G, 1, $Nodes[$I][1] );
      }
   }
   close MACRO;

   if ( $Ex == 1 ) {
      system "dsmadmc $Id $Pa -noconfirm -itemcommit macro $Macro";
      unlink "$Macro";
   }
}
